(function(){
  var ElementUtil = function() {};
  var MAX_WIDTH = 360;

  ElementUtil.instance = function() {
    if (!!!ElementUtil.$instance) {
      ElementUtil.$instance = new ElementUtil;
    }
    return ElementUtil.$instance;
  };

  ElementUtil.prototype.getBoundingRect = function(e) {
    var box = e.getBoundingClientRect();
    return {
      top: box.top + (window.pageYOffset || document.documentElement.scrollTop) - (document.documentElement.clientTop || 0),
      left: box.left + (window.pageXOffset || document.documentElement.scrollLeft) - (document.documentElement.clientLeft || 0),
      width: box.right - box.left,
      height: box.bottom - box.top
    }
  };

  ElementUtil.prototype.getFloatValue = function(val) {
    return parseFloat(val) || 0;
  }

  ElementUtil.prototype.getPadding = function(style) {
    var left = style.getPropertyValue('padding-left');
    var right = style.getPropertyValue('padding-right');
    var top = style.getPropertyValue('padding-top');
    var bottom = style.getPropertyValue('padding-bottom');
    return {
      left: !!left ? this.getFloatValue(left) : 0,
      right: !!right ? this.getFloatValue(right) : 0,
      top: !!top ? this.getFloatValue(top) : 0,
      bottom: !!bottom ? this.getFloatValue(bottom) : 0
    };
  };

  ElementUtil.prototype.maxPhotoWidth = function() {
    return MAX_WIDTH;
  };

  ElementUtil.prototype.applyEdgeToEdgeStyle = function(img) {
    var p = img.ownerDocument.createElement('p');
    var originWidth = img.width;
    var originHeight = img.height;
    var width = this.maxPhotoWidth();
    var scale = width / originWidth;
    var height = originHeight * scale;
    p.style.setProperty('height', height + 'px');
    p.style.setProperty('margin-top', '0px');
    p.style.setProperty('margin-bottom', '1px');
    img.parentElement.replaceChild(p, img);
    p.appendChild(img);
    img.style.removeProperty('width');
    img.style.removeProperty('height');
    // mf_edgeToEdgeImageAttachmentStyle begin
    img.style.setProperty('left', '0');
    img.style.setProperty('position', 'absolute');
    img.style.setProperty('right', '0');
    img.style.setProperty('width', '100%');
    // mf_edgeToEdgeImageAttachmentStyle end
    img.removeAttribute('WIDTH');
    img.removeAttribute('HEIGHT');
  };

  ElementUtil.prototype.isEdgeToEdgeFormatted = function(img) {
    var parentElement = img.parentElement;
    if (parentElement.nodeName.toUpperCase() === 'P') {
      if (img.style.getPropertyValue('width') === '100%') {
        if (img.style.getPropertyValue('position') === 'absolute') {
          if (!!!this.isContainedInTable(img)) {
            return true;
          }
        }
      }
    }
    return false;
  };

  ElementUtil.prototype.shouldFormatEdgeToEdge = function(img) {
    var width = img.ownerDocument.defaultView.getComputedStyle(img).getPropertyValue('width').floatValue;
    if (width >= this.maxPhotoWidth()) {
      if (!!!this.isEdgeToEdgeFormatted(img) && !!!this.isContainedInQuoteBlock(img) && !!!this.isContainedInTable(img)) {
        return true;
      }
    }
    return false;
  };

  ElementUtil.prototype.isContainedInTable = function(e) {
    for (var p = e.parentElement; !!p; p = p.parentElement) {
      if (p.tagName.toUpperCase() === 'TABLE') {
        return true;
      }
    }
    return false;
  };

  ElementUtil.prototype.isContainedInQuoteBlock = function(e) {
    return !!this.containingQuoteBlockElement(e);
  };

  ElementUtil.prototype.containingQuoteBlockElement = function(e) {
    for (var p = e.parentElement; !!p; p = p.parentElement) {
      if (p.tagName.toUpperCase() === 'BLOCKQUOTE' && p.getAttribute('TYPE').toUpperCase() === 'CITE') {
        return p;
      }
    }
    return null;
  };

  ElementUtil.prototype.isContainedInAnyElementInArray = function(e, a) {
    if (!!!a || a.length === 0) {
      return false;
    }
    for (; !!e; e = e.parentElement) {
      if (a.indexOf(e) !== -1) {
        return true;
      }
    }
    return false;
  };

  var MessageReformattingContext = function(document) {
    this.document = document;
    this.body = document.body;
    this.changes = [];
    this._rightFloats = [];
    this._hasAnyLeftFloat = false;
    this.floatsNeedUpdate = true;
    this.metricsNeedUpdate = true;
    this.minimumRescalingFactor = 1.0;
    this._maximumWidth = 0;
    this._meanWidth = 0;
    this._deviationWidth = 0;
    this.documentContainsAnyWebKitTransform = false;
  };

  MessageReformattingContext.prototype.updateFloatsIfNecessary = function() {
    if (!!this.floatsNeedUpdate) {
      var it = this.document.createNodeIterator(this.body, 1, 0, 1);
      for (var n = it.nextNode(); !!n; n = it.nextNode()) {
        if (!(n instanceof Element) || !n.hasChildNodes()) {
          continue;
        }
        var floatCssText = this.document.defaultView.getComputedStyle(n).getPropertyValue('float').toUpperCase();
        if (floatCssText === 'RIGHT') {
          this._rightFloats.push(n);
        } else if (floatCssText === 'LEFT'){
          this._hasAnyLeftFloat = true;
        }
      }
      this.floatsNeedUpdate = false;
    }
  };

  MessageReformattingContext.prototype.changeSizeOfElement = function(e, priority, block) {
    var style = this.document.defaultView.getComputedStyle(e);
    var width = ElementUtil.instance().getFloatValue(style.getPropertyValue('width'));
    var height = ElementUtil.instance().getFloatValue(style.getPropertyValue('height'));
    var paddings = ElementUtil.instance().getPadding(style);
    var resizeResult = block(width, height, paddings.top, paddings.left, paddings.bottom, paddings.right);
    // resizeResult -> {width:, height:, shouldNotResize:};
    if (!!!resizeResult.shouldNotResize) {
      var newWidth = resizeResult.width;
      var newHeight = resizeResult.height;
      if (newWidth != 0 || newHeight != 0) {    // almost equal?
        var change = {element: e};

        if (newHeight > 0.0 && e.style.height != "") {
          change['height'] = height;
        }
        if (newWidth > 0.0 && e.style.width != "") {
          change['width'] = width;
        }
        this.changes.push(change);
        if (newHeight > 0) {
          e.style.setProperty('height', newHeight+'px', 'important');
        }
        if (newWidth > 0) {
          e.style.setProperty('width', newWidth+'px', 'important');
        }
        this.metricsNeedUpdate = true;
        return true;
      }
    }
    return false;
  };

  MessageReformattingContext.prototype.resizeElement = function(e, scale, verificationBlock) {
    return this.changeSizeOfElement(e, 'important', function(w, h, pTop, pLeft, pBottom, pRight) {
      var resultW = w;
      var resultH = 0;
      var pX = pLeft + pRight;
      resultW += pX;
      resultW *= scale;
      resultW -= pX;
      if (e instanceof HTMLImageElement) {
        var pY = pTop + pBottom;
        resultH = h;
        resultH += pY;
        resultH *= scale;
        resultH -= pY;
      }
      var shouldNotResize = false;
      if (!!verificationBlock) {
        if (!!!verificationBlock(w, h, resultW, resultH)) {
          shouldNotResize = true;
        }
      }
      return {
        width: resultW,
        height: resultH,
        shouldNotResize: shouldNotResize
      };
    });
  };

  MessageReformattingContext.prototype.hasAnyLeftFloat = function() {
    this.updateFloatsIfNecessary();
    return this._hasAnyLeftFloat;
  };

  MessageReformattingContext.prototype.hasAnyRightFloat = function() {
    this.updateFloatsIfNecessary();
    return this._rightFloats.size > 0;
  };

  MessageReformattingContext.prototype.rightFloats = function() {
    this.updateFloatsIfNecessary();
    return this._rightFloats.slice(0);    // copy
  };

  MessageReformattingContext.prototype.rollBackLastChangeForElement = function(e) {
    if (this.changes.size == 0) {
      alert('warning');
    }
    var change = this.changes[this.changes.length-1];
    if (!(change['element'] === e)) {
      alert('warning');
    }
    this.rollBackChange(change);
    this.changes.splice(this.changes.length-1);
  };

  MessageReformattingContext.prototype.rollBackChange = function(change) {
    var e = change['element'];
    var h = change['height'];
    var w = change['width'];
    if (!!h) {
      if (!!w) {
        e.style.setProperty('height', h+'px', 'important');
      } else {
        e.style.removeProperty('height');
      }
      this.metricsNeedUpdate = true;
    }
    if (!!w) {
      e.style.setProperty('width', w+'px', 'important');
      this.metricsNeedUpdate = true;
    }
    var m = change['margin'];
    if (!!m) {
      e.style.setProperty('margin', m, 'important');
      this.metricsNeedUpdate = true;
    } else {
      e.style.removeProperty('margin');
    }
    var property = change['property'];
    if (!!property) {
      var v = change['value'];
      if (!!v) {
        e.style.setProperty(property, v, change['priority']);
      } else {
        e.style.removeProperty(property);
      }
      this.metricsNeedUpdate = true;
    }
  };

  MessageReformattingContext.prototype.rollBackAllChanges = function() {
    for (var i = this.changes.length-1; i >= 0; --i) {
      this.rollBackChange(this.changes[i]);
    }
    this.changes = [];
    this.minimumRescalingFactor = 1.0;
  }


  MessageReformattingContext.prototype.rescaleElement = function(e, scale) {
    var d = e.ownerDocument;
    var p = e.parentElement;
    var result = false;
    if (!(p === d.documentElement)) {
      var rect = ElementUtil.instance().getBoundingRect(e);
      var width = rect.width;
      var height = rect.height;
      this.changeSizeOfElement(e, 'important', function(){
        return {width:width, height:height, shouldNotResize:false};
      });
      this.changeProperty(e, '-webkit-transform-origin', '0 0', 'important');
      this.changeProperty(e, '-webkit-transform', 'scale('+scale+')', 'important');
      result = true;
      this.documentContainsAnyWebKitTransform = true;
      var remainingWidth = width * (1 - scale);
      var remainingHeight = height * (1 - scale);
      this.changeMarginOfElement(e, 'important', function(originPadding) {
        return {
          top: originPadding.top * scale,
          left: originPadding.left * scale,
          bottom: originPadding.bottom * scale - remainingHeight,
          right: originPadding.right * scale - remainingWidth
        };
      });
      if (e != document.documentElement) {
        var range = d.createRange();
        range.setStartBefore(e);
        range.setEndAfter(e);
        var div = d.createElement('div');
        range.surroundContents(div);
        this.changeProperty(div, 'height', (height * scale)+'px', 'important');
      }

      if (this.minimumRescalingFactor > scale) {
        this.minimumRescalingFactor = scale;
      }
    }
    return result;
  };

  MessageReformattingContext.prototype.changeProperty = function(e, property, v, priority) {
    var change = {'element':e, 'property':property};
    var style = e.style;
    var originValue = e.style.getPropertyValue(property);
    var originPriority = e.style.getPropertyPriority(property);
    if (!!originValue) {
      change['value'] = originValue;
      if (!!originPriority) {
        change['priority'] = originPriority;
      }
    }
    this.changes.push(change);
    style.setProperty(property, v, priority);
    this.metricsNeedUpdate = true;
  };

  MessageReformattingContext.prototype.changeMarginOfElement = function(e, priority, block) {
    var style = this.document.defaultView.getComputedStyle(e);
    var paddings = {
      top: 0,
      left: 0,
      bottom: 0,
      right: 0
    };
    if (!!style) {
      paddings = ElementUtil.instance().getPadding(style);
    }
    var blockResult = block(paddings);
    // blockResult -> {top:,left:,bottm:,right:,shouldNotChange:}
    if (!!!blockResult.shouldNotChange) {
      var change = {'element': e};
      change['margin'] = paddings.top+'px '+paddings.right+'px '+ paddings.bottom+'px '+paddings.left+'px';
      // 如果没有margin定义则rollback时直接移除
      if ("" == e.style.margin) {
        change['margin'] = undefined;
      }
      this.changes.push(change)
      e.style.setProperty('margin', blockResult.top+'px '+blockResult.right+'px '+ blockResult.bottom+'px '+blockResult.left+'px');
      this.metricsNeedUpdate = true;
    }
    return !!!blockResult.shouldNotChnage;
  };

  MessageReformattingContext.prototype.meanWidth = function() {
    this.updateMetricsIfNecessary();
    return this._meanWidth;
  };

  MessageReformattingContext.prototype.maximumWidth = function() {
    this.updateMetricsIfNecessary();
    return this._maximumWidth;
  };

  MessageReformattingContext.prototype.deviationWidth = function() {
    this.updateMetricsIfNecessary();
    return this._deviationWidth;
  }

  MessageReformattingContext.prototype.updateMetricsIfNecessary = function() {
    if (!!this.metricsNeedUpdate) {
      var it = this.document.createNodeIterator(this.body, 1, 0, 1);
      var widths = [];
      var totalWidth = 0;
      var maxWidth = 0;
      for(var n = it.nextNode(); !!n; n = it.nextNode()) {
        var rect = ElementUtil.instance().getBoundingRect(n);
        var width = rect.width;
        if (width <= 10.0) {
          continue;
        }
        widths.push(width);
        totalWidth += width;
        var right = rect.left + width;
        if (right > maxWidth) {
          maxWidth = right;
        }
      }
      var count = widths.length;
      var meanWidth = totalWidth / count;
      var squareDeviationWidth = 0;
      for (var i = 0; i < count; ++i) {
        var wDiff = widths[i] - meanWidth;
        squareDeviationWidth += wDiff * wDiff;
      }
      this._maximumWidth = maxWidth;
      this._meanWidth = meanWidth;
      this._deviationWidth = Math.sqrt(squareDeviationWidth / count);
      this.metricsNeedUpdate = false;
    }
  };

  MessageReformattingContext.prototype.didChangeDocument = function() {
    return this.changes.length > 0;
  };

  var ReformattingCore = function() {
    this.context = new MessageReformattingContext(document);
  };

  ReformattingCore.prototype.setWidth = function(width) {
    this.defWidth = width;
    MAX_WIDTH = width;
  }

  ReformattingCore.prototype.elementHasDefinedWidth = function(e) {
    var style = e.ownerDocument.defaultView.getComputedStyle(e);
    var widthStyle = style.getPropertyValue('width');
    if (-1 == widthStyle.indexOf('px')) {
      return false;
    }
    return !(ElementUtil.instance().getFloatValue(widthStyle) === 0);
  }

  ReformattingCore.prototype.defaultWidth = function() {
    return this.defWidth;
  }

  ReformattingCore.prototype.reformatOneElement = function(block) {
    var formatEdgeToEdgeArray = [];
    var it = this.context.document.createNodeIterator(this.context.body, 1, 0, 1);
    var suitableDefaultWidth = this.defaultWidth() - 20;
    var recordedNode = null;
    var recordedWidth = 0;
    var recordedNodeHasDefinedWidth = null;
    for(var n = it.nextNode(); !!n; n = it.nextNode()) {
      var shouldFormatEdgeToEdge = false;
      if ((n instanceof HTMLImageElement) && ElementUtil.instance().shouldFormatEdgeToEdge(n)) {
        shouldFormatEdgeToEdge = true;
      }
      if (!(n instanceof HTMLAreaElement) && !shouldFormatEdgeToEdge) {
        var rect = ElementUtil.instance().getBoundingRect(n);
        var needRecordFlag = false;
        var width = rect.width;
        if (width > suitableDefaultWidth && (n instanceof HTMLElement)) {
          needRecordFlag = true;
        }
        var needRecordFlag2 = false;
        if (width >= recordedWidth) {
          if (width > recordedWidth) {
            recordedWidth = width;
            needRecordFlag2 = true;
          } else if (needRecordFlag) {
            var elementHasDefinedWidth = this.elementHasDefinedWidth(n);
            if (elementHasDefinedWidth) {
              if (recordedNodeHasDefinedWidth === null) {
                recordedNodeHasDefinedWidth = this.elementHasDefinedWidth(recordedNode);
              }
              if (!!recordedNodeHasDefinedWidth) {
                needRecordFlag = false;
              }
              needRecordFlag2 = true;
            }
          }
          if (!!needRecordFlag && !!needRecordFlag2) {
            recordedNodeHasDefinedWidth = null;
            recordedNode = n;
          }
        }
      }
    }

    var formatEdgeToEdgeCount = formatEdgeToEdgeArray.lenth;
    if (formatEdgeToEdgeCount > 0) {
      for (var i = 0; i < formatEdgeToEdgeCount; ++i) {
        ElementUtil.instance().applyEdgeToEdgeStyle(formatEdgeToEdgeArray[i]);
      }
      if (!!!recordedNode) {
        var width = ElementUtil.instance().getBoundingRect(recordedNode).width;
        if (width < recordedWidth) {
          recordedWidth = width;
        }
      }
    }

    if (recordedNode instanceof HTMLImageElement) {
      if (ElementUtil.instance().isEdgeToEdgeFormatted(recordedNode)) {
        recordedNode = null;
      }
    }

    var shouldCancel = false;
    var resultRecordedNode = null;
    if (!!recordedNode) {
      var rect = ElementUtil.instance().getBoundingRect(recordedNode);
      if (rect.left <= suitableDefaultWidth * 0.5) {
        var shouldResize = false;
        if (!!block) {
          var blockResult = block(recordedNode);
          if (blockResult == 2) {
            var ratio = suitableDefaultWidth / rect.width;
            if (this.context.rescaleElement(recordedNode, ratio)) {
              resultRecordedNode = recordedNode;
            }
          } else if (blockResult == 1) {
            shouldResize = true;
          }
        } else {
          shouldResize = true;
        }
        if (shouldResize) {
          var ratio = suitableDefaultWidth / rect.width;
          var resizeResult = this.context.resizeElement(recordedNode, ratio, function(width) {
            return width > suitableDefaultWidth;
          });
          if (!!resizeResult) {
            resultRecordedNode = recordedNode;
          }
        }
      } else {
        shouldCancel = true;
      }
    }
    return {
      'element': resultRecordedNode,
      'shouldCancel' : shouldCancel
    }
  };

  ReformattingCore.prototype.rescaleTopLevelElements = function() {
    if (!!!this.context.hasAnyLeftFloat()) {
      var rightFloats = this.context.rightFloats();
      var quotedArray = new Array();      // ios 8 才有Set
      if (rightFloats.length != 0) {
        for (var e in rightFloats) {
          var e = e.parentElement;
          for ( ; !!e && ElementUtil.instance().isContainedInAnyElementInArray(e, rightFloats); e = e.parentElement)
            ;
          if (!!!e) {
            return false;
          }
          var quotedElement = ElementUtil.instance().containingQuoteBlockElement(e);
          if (!!!quotedElement) {
            return false;
          }
          if (quotedArray.indexOf(quotedElement) === -1) {
            quotedArray.push(quotedElement);
          }
        }
      }

      var it = this.context.document.createNodeIterator(this.context.body, 1, 0, 1);
      var defaultWidth = this.defaultWidth();
      var needRescaledNodes = [];
      var hasPreContainedInTable_var_12c = false;
      var hasImgNotContainedInTable_var_134 = false;
      var hasNeedRescaled_var_138 = false;
      for (var n = it.nextNode(); !!n; n = it.nextNode()) {
        if (!(n instanceof Element) || ElementUtil.instance().isContainedInAnyElementInArray(n, quotedArray)) {
          continue;
        }
        var tagName = n.tagName.toUpperCase();
        if (tagName === 'IMG') {
          if (!ElementUtil.instance().isContainedInTable(n)) {
            hasImgNotContainedInTable_var_134 = true;
            if (ElementUtil.instance().getBoundingRect(n).left > defaultWidth * 0.25) {
              return false;
            }
          }
        } else if (tagName === 'PRE') {
          if (ElementUtil.instance().isContainedInTable(n)) {
            hasPreContainedInTable_var_12c = true;
          }
        } else if (tagName === 'TABLE') {
          if (!ElementUtil.instance().isContainedInTable(n)) {
            // TODO:此处少一个根据text计算是不是要fail的逻辑，关键代码：[self _rangeOfFirstText]
            hasNeedRescaled_var_138 = true;
            needRescaledNodes.push(n);
          }
        }
      }

      var result = true;
      if (!!!hasNeedRescaled_var_138) {
        if (quotedArray.length == 0 && !hasImgNotContainedInTable_var_134) {
          result = hasPreContainedInTable_var_12c;
          if (!result) {
            return false;
          }
        }
      }

      var defaultWidth = this.defaultWidth();
      var suitableDefaultWidth = defaultWidth;
      if (needRescaledNodes.length != 0) {
        suitableDefaultWidth -= 20;
      }
      for (var i = needRescaledNodes.length-1; i >= 0; --i) {
        var n = needRescaledNodes[i];
        var rect = ElementUtil.instance().getBoundingRect(n);
        var right = rect.left + rect.width;
        var scale = suitableDefaultWidth / right;
        if (!!!this.context.rescaleElement(n, scale)) {
          return false;
        }
      }
      return result;
    }
    return false;
  }

  ReformattingCore.prototype.reformatMessage = function() {
    // TODO: 暂时忽略了skipResize和skipReformat
    var maximumWidth = this.context.maximumWidth();
    var defaultWidth = this.defaultWidth();
    if (maximumWidth > defaultWidth) {
      // resize
      if (!this.context.hasAnyRightFloat()) {
        if (this.shouldResizeMessage()) {
          var shouldResizeMessageAndEqualLastReformatElement = false;
          var remformatFailed = false;
          var shouldCancel = false;
          var lastReformatedElement = null;
          var currentResizeElement = null;
          for (;;) {
            var shouldResizeMessage = this.shouldResizeMessage();
            if (!shouldResizeMessage && (!!!lastReformatedElement || !this.shouldContinueResizingMessage())) {
              break;
            }
            var reformatResult = this.reformatOneElement();
            currentResizeElement = reformatResult.element;
            shouldCancel = reformatResult.shouldCancel;
            if (currentResizeElement === lastReformatedElement) {
              shouldResizeMessageAndEqualLastReformatElement = true;
              break;
            }
            lastReformatedElement = currentResizeElement;
            if (!!shouldCancel) {
              remformatFailed = true;
              break;
            }
          }
          if (!reformattingDidFail) {
            if (!!shouldCancel) {
              reformattingDidFail = true;
            } else if (!!lastReformatedElement) {
              if (shouldResizeMessageAndEqualLastReformatElement && shouldResizeMessage) {
                reformattingDidFail = true;
              }
            }
          }
          if (reformattingDidFail) {
            this.reformattingDidFail();
          }
        }
      }
      // rescale
      if (this.shouldRescaleMessage()) {
        if (this.rescaleTopLevelElements()) {
          var lastReformatedElement = null;
          var rescaledElement = null;
          for (;;) {
            var tryReformated = false;
            var reformatMode = 0;
            var that = this;
            var reformatResult = this.reformatOneElement(function(e) {
              tryReformated = true;
              if (rescaledElement != e) {
                if (lastReformatedElement != e) {
                  reformatMode = 1;   // resize
                } else {
                  that.context.rollBackLastChangeForElement(e);
                  reformatMode = 2;   // rescale;
                }
              } else {
                reformatMode = 0;     // nothing
              }
              return reformatMode;
            });
            var currentReformatedElement = reformatResult.element;
            var shouldCancel = reformatResult.shouldCancel;
            var reformattingDidFail = false;
            if (shouldCancel) {
              reformattingDidFail = true;
            } else if (!!!currentReformatedElement){
              if (!!tryReformated && reformatMode == 2) {
                reformattingDidFail = true;
              } else {
                break;      // 正常结束
              }
            } else if (currentReformatedElement === rescaledElement) {
              reformattingDidFail = true;
            } else {
              if (currentReformatedElement === lastReformatedElement) {
                rescaledElement = currentReformatedElement;
              }
              lastReformatedElement = currentReformatedElement;
            }
            if (reformattingDidFail) {
              this.reformattingDidFail();
              break;
            }
          }
        }
      }
    }

    // 邮箱大师附加逻辑：如果最后发现HTML的元素还是宽与屏幕，则做documentElement上的scale
    // 但为了防止大量元素堆积造成的宽度过大，因此此处是取maxWidth，而不是documentElement.scrollWidth
   if (this.defaultWidth() < this.context.maximumWidth() ) {
     var scale1 = this.defaultWidth() / this.context.maximumWidth();
     var scale2 = Math.min(1, document.documentElement.clientHeight / document.documentElement.scrollHeight);
     // do not use this in android. for adjust attchments position.
     //this.context.rescaleElement(document.documentElement, Math.max(scale1, scale2));
     //this.scale = Math.max(scale1, scale2);
   }
  }

  ReformattingCore.prototype.getDocumentScale = function() {
    return this.scale;
  }

  ReformattingCore.prototype.reformattingDidFail = function() {
    this.context.rollBackAllChanges();
  }

  ReformattingCore.prototype.shouldResizeMessage = function() {
    var deviationWidth = this.context.deviationWidth();
    var meanWidth = this.context.meanWidth();
    if (deviationWidth > meanWidth) {
      return true;
    }
    var maximumWidth = this.context.maximumWidth();
    if (Math.abs(maximumWidth - deviationWidth) < deviationWidth * 1.1) {
      return true;
    }
    return false;
  };

  ReformattingCore.prototype.shouldContinueResizingMessage = function() {
    if (this.context.maximumWidth() > this.defaultWidth()) {
      return true;
    }
    return false;
  };

  ReformattingCore.prototype.shouldRescaleMessage = function() {
    if ((this.defaultWidth() - 20) / this.context.maximumWidth() < 0.75) {
      return true;
    }
    return false;
  };

  window.ElementUtil = ElementUtil;
  window.MessageReformattingContext = MessageReformattingContext;
  window.ReformattingCore = ReformattingCore;
})();